/*
 * Copyright (C) 2012 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Author: Olivier Fourdan <ofourdan@redhat.com>
 *
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <cairo.h>

#include "gsd-wacom-osd-window.h"
#include "gsd-wacom-device.h"
#include "gsd-enums.h"

#define ROTATION_KEY                "rotation"
#define ACTION_TYPE_KEY             "action-type"
#define CUSTOM_ACTION_KEY           "custom-action"
#define CUSTOM_ELEVATOR_ACTION_KEY  "custom-elevator-action"

#define BUTTON_ASPECT		0.8
#define BUTTON_MAX_WIDTH	120
#define BUTTON_SPACING		30
#define LABEL_SPACING		20
#define LINE_WIDTH		5
#define WINDOW_OPACITY		0.8

static void
draw_rounded_rectangle (cairo_t         *cr,
			double           x,
			double           y,
			double           width,
			double           height)
{
	double radius = height / 10.0;
	double degrees = G_PI / 180.0;

	cairo_new_sub_path (cr);
	cairo_arc (cr, x + width - radius, y + radius, radius, -90 * degrees, 0 * degrees);
	cairo_arc (cr, x + width - radius, y + height - radius, radius, 0 * degrees, 90 * degrees);
	cairo_arc (cr, x + radius, y + height - radius, radius, 90 * degrees, 180 * degrees);
	cairo_arc (cr, x + radius, y + radius, radius, 180 * degrees, 270 * degrees);
	cairo_close_path (cr);
}

static void
draw_circle (cairo_t         *cr,
	     double           x,
	     double           y,
	     double           radius)
{
	cairo_new_sub_path (cr);
	cairo_arc (cr, x, y, radius, 0, 2 * G_PI);
	cairo_close_path (cr);
}

#define GSD_TYPE_WACOM_OSD_BUTTON         (gsd_wacom_osd_button_get_type ())
#define GSD_WACOM_OSD_BUTTON(o)           (G_TYPE_CHECK_INSTANCE_CAST ((o), GSD_TYPE_WACOM_OSD_BUTTON, GsdWacomOSDButton))
#define GSD_WACOM_OSD_BUTTON_CLASS(k)     (G_TYPE_CHECK_CLASS_CAST((k), GSD_TYPE_WACOM_OSD_BUTTON, GsdWacomOSDButtonClass))
#define GSD_IS_WACOM_OSD_BUTTON(o)        (G_TYPE_CHECK_INSTANCE_TYPE ((o), GSD_TYPE_WACOM_OSD_BUTTON))
#define GSD_IS_WACOM_OSD_BUTTON_CLASS(k)  (G_TYPE_CHECK_CLASS_TYPE ((k), GSD_TYPE_WACOM_OSD_BUTTON))
#define GSD_WACOM_OSD_BUTTON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GSD_TYPE_WACOM_OSD_BUTTON, GsdWacomOSDButtonClass))

typedef struct GsdWacomOSDButtonPrivate GsdWacomOSDButtonPrivate;

typedef struct {
        GObject                   parent;
        GsdWacomOSDButtonPrivate *priv;
} GsdWacomOSDButton;

typedef struct {
        GObjectClass              parent_class;
} GsdWacomOSDButtonClass;

GType                     gsd_wacom_osd_button_get_type        (void) G_GNUC_CONST;

enum {
	PROP_OSD_BUTTON_0,
	PROP_OSD_BUTTON_ID,
	PROP_OSD_BUTTON_LABEL,
	PROP_OSD_BUTTON_ACTIVE
};

#define GSD_WACOM_OSD_BUTTON_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), \
					     GSD_TYPE_WACOM_OSD_BUTTON, \
					     GsdWacomOSDButtonPrivate))

struct GsdWacomOSDButtonPrivate {
	char                     *id;
	char                     *label;
	GdkRectangle              area;
	GsdWacomTabletButtonType  type;
	GsdWacomTabletButtonPos   position;
	gboolean                  active;
};

static void     gsd_wacom_osd_button_class_init  (GsdWacomOSDButtonClass *klass);
static void     gsd_wacom_osd_button_init        (GsdWacomOSDButton      *osd_button);
static void     gsd_wacom_osd_button_finalize    (GObject                *object);

G_DEFINE_TYPE (GsdWacomOSDButton, gsd_wacom_osd_button, G_TYPE_OBJECT)

static void
gsd_wacom_osd_button_set_id (GsdWacomOSDButton *osd_button,
			     const gchar       *str)
{
	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON (osd_button));
	g_return_if_fail (str != NULL);

	g_free (osd_button->priv->id);
	osd_button->priv->id = g_strdup (str);
}

static void
gsd_wacom_osd_button_set_label (GsdWacomOSDButton *osd_button,
				const gchar       *str)
{
	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON (osd_button));

	g_free (osd_button->priv->label);
	osd_button->priv->label = g_strdup (str ? str : "");
}

static void
gsd_wacom_osd_button_set_button_type (GsdWacomOSDButton        *osd_button,
				      GsdWacomTabletButtonType  type)
{
	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON (osd_button));

	osd_button->priv->type = type;
}

static void
gsd_wacom_osd_button_set_position (GsdWacomOSDButton        *osd_button,
				   GsdWacomTabletButtonPos   position)
{
	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON (osd_button));

	osd_button->priv->position = position;
}

static void
gsd_wacom_osd_button_set_active (GsdWacomOSDButton *osd_button,
				 gboolean           active)
{
	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON (osd_button));

	osd_button->priv->active = active;
}

static void
gsd_wacom_osd_button_move (GsdWacomOSDButton *osd_button,
			   gint               x,
			   gint               y)
{
	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON (osd_button));

	osd_button->priv->area.x = x;
	osd_button->priv->area.y = y;
}

static void
gsd_wacom_osd_button_resize (GsdWacomOSDButton *osd_button,
			     gint               width,
			     gint               height)
{
	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON (osd_button));

	osd_button->priv->area.width = width;
	osd_button->priv->area.height = height;
}

static GsdWacomOSDButton *
gsd_wacom_osd_button_new (gint          x,
			  gint          y,
			  gint          width,
			  gint          height)
{
	GsdWacomOSDButton *osd_button;

	osd_button = GSD_WACOM_OSD_BUTTON (g_object_new (GSD_TYPE_WACOM_OSD_BUTTON, NULL));
	osd_button->priv->area.x = x;
	osd_button->priv->area.y = y;
	osd_button->priv->area.width = width;
	osd_button->priv->area.height = height;

	return osd_button;
}

static void
gsd_wacom_osd_button_set_property (GObject        *object,
				   guint           prop_id,
				   const GValue   *value,
				   GParamSpec     *pspec)
{
	GsdWacomOSDButton *osd_button;

	osd_button = GSD_WACOM_OSD_BUTTON (object);

	switch (prop_id) {
	case PROP_OSD_BUTTON_ID:
		gsd_wacom_osd_button_set_id (osd_button, g_value_get_string (value));
		break;
	case PROP_OSD_BUTTON_LABEL:
		gsd_wacom_osd_button_set_label (osd_button, g_value_get_string (value));
		break;
	case PROP_OSD_BUTTON_ACTIVE:
		gsd_wacom_osd_button_set_active (osd_button, g_value_get_boolean (value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
gsd_wacom_osd_button_get_property (GObject        *object,
				   guint           prop_id,
				   GValue         *value,
				   GParamSpec     *pspec)
{
	GsdWacomOSDButton *osd_button;

	osd_button = GSD_WACOM_OSD_BUTTON (object);

	switch (prop_id) {
	case PROP_OSD_BUTTON_ID:
		g_value_set_string (value, osd_button->priv->id);
		break;
	case PROP_OSD_BUTTON_LABEL:
		g_value_set_string (value, osd_button->priv->label);
		break;
	case PROP_OSD_BUTTON_ACTIVE:
		g_value_set_boolean (value, osd_button->priv->active);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
gsd_wacom_osd_button_class_init (GsdWacomOSDButtonClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->set_property = gsd_wacom_osd_button_set_property;
	object_class->get_property = gsd_wacom_osd_button_get_property;
	object_class->finalize = gsd_wacom_osd_button_finalize;

	g_object_class_install_property (object_class,
	                                 PROP_OSD_BUTTON_ID,
	                                 g_param_spec_string ("id",
	                                                      "Button Id",
	                                                      "The Wacom Button ID",
	                                                      "",
	                                                      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
	                                 PROP_OSD_BUTTON_LABEL,
	                                 g_param_spec_string ("label",
	                                                      "Label",
	                                                      "The button label",
	                                                      "",
	                                                      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
	                                 PROP_OSD_BUTTON_ACTIVE,
	                                 g_param_spec_boolean ("active",
	                                                       "Active",
	                                                       "Whether the button is active",
	                                                       FALSE,
	                                                       G_PARAM_READWRITE));

	g_type_class_add_private (klass, sizeof (GsdWacomOSDButtonPrivate));
}

static void
gsd_wacom_osd_button_init (GsdWacomOSDButton *osd_button)
{
	osd_button->priv = GSD_WACOM_OSD_BUTTON_GET_PRIVATE (osd_button);
}

static void
gsd_wacom_osd_button_finalize (GObject *object)
{
	GsdWacomOSDButton *osd_button;
	GsdWacomOSDButtonPrivate *priv;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON (object));

	osd_button = GSD_WACOM_OSD_BUTTON (object);

	g_return_if_fail (osd_button->priv != NULL);

	priv = osd_button->priv;

	g_free (priv->id);
	priv->id = NULL;

	g_free (priv->label);
	priv->label = NULL;

	G_OBJECT_CLASS (gsd_wacom_osd_button_parent_class)->finalize (object);
}

/* Compute the new actual position once rotation is applied */
static GsdWacomTabletButtonPos
get_actual_position (GsdWacomTabletButtonPos position,
		     GsdWacomRotation        rotation)
{
	switch (rotation) {
	case GSD_WACOM_ROTATION_NONE:
		return position;
		break;
	case GSD_WACOM_ROTATION_HALF:
		if (position == WACOM_TABLET_BUTTON_POS_LEFT)
			return WACOM_TABLET_BUTTON_POS_RIGHT;
		if (position == WACOM_TABLET_BUTTON_POS_RIGHT)
			return WACOM_TABLET_BUTTON_POS_LEFT;
		if (position == WACOM_TABLET_BUTTON_POS_TOP)
			return WACOM_TABLET_BUTTON_POS_BOTTOM;
		if (position == WACOM_TABLET_BUTTON_POS_BOTTOM)
			return WACOM_TABLET_BUTTON_POS_TOP;
		break;
	case GSD_WACOM_ROTATION_CCW:
		if (position == WACOM_TABLET_BUTTON_POS_LEFT)
			return WACOM_TABLET_BUTTON_POS_BOTTOM;
		if (position == WACOM_TABLET_BUTTON_POS_RIGHT)
			return WACOM_TABLET_BUTTON_POS_TOP;
		if (position == WACOM_TABLET_BUTTON_POS_TOP)
			return WACOM_TABLET_BUTTON_POS_LEFT;
		if (position == WACOM_TABLET_BUTTON_POS_BOTTOM)
			return WACOM_TABLET_BUTTON_POS_RIGHT;
		break;
	case GSD_WACOM_ROTATION_CW:
		if (position == WACOM_TABLET_BUTTON_POS_LEFT)
			return WACOM_TABLET_BUTTON_POS_TOP;
		if (position == WACOM_TABLET_BUTTON_POS_RIGHT)
			return WACOM_TABLET_BUTTON_POS_BOTTOM;
		if (position == WACOM_TABLET_BUTTON_POS_TOP)
			return WACOM_TABLET_BUTTON_POS_RIGHT;
		if (position == WACOM_TABLET_BUTTON_POS_BOTTOM)
			return WACOM_TABLET_BUTTON_POS_LEFT;
		break;
	default:
		break;
	}
	/* fallback, should not happen */
	return position;
}

/* Compute the new actual index once rotation is applied */
static gint
get_actual_index (GsdWacomTabletButtonPos position,
		  GsdWacomRotation        rotation,
		  gint                    n_items,
		  gint                    current)
{
	g_return_val_if_fail (current < n_items, current);

	switch (rotation) {
	case GSD_WACOM_ROTATION_NONE:
			return current;
			break;
		case GSD_WACOM_ROTATION_HALF:
			return n_items - current - 1;
			break;
		case GSD_WACOM_ROTATION_CCW:
			if (position == WACOM_TABLET_BUTTON_POS_LEFT ||
			    position == WACOM_TABLET_BUTTON_POS_RIGHT)
				return current;
			if (position == WACOM_TABLET_BUTTON_POS_TOP ||
			    position == WACOM_TABLET_BUTTON_POS_BOTTOM)
				return n_items - current - 1;
			break;
		case GSD_WACOM_ROTATION_CW:
			if (position == WACOM_TABLET_BUTTON_POS_LEFT ||
			    position == WACOM_TABLET_BUTTON_POS_RIGHT)
				return n_items - current - 1;
			if (position == WACOM_TABLET_BUTTON_POS_TOP ||
			    position == WACOM_TABLET_BUTTON_POS_BOTTOM)
				return current;
			break;
		default:
			break;
	}
	/* fallback, should not happen */
	return current;
}

static void
gsd_wacom_osd_button_draw (GsdWacomOSDButton *osd_button,
			   GtkStyleContext   *style_context,
			   PangoContext      *pango_context,
			   cairo_t           *cr,
			   GtkAllocation     *allocation,
			   GsdWacomRotation   rotation)
{
	GsdWacomOSDButtonPrivate *priv;
	PangoLayout              *layout;
	PangoRectangle            logical_rect;
	GsdWacomTabletButtonPos   actual_position;
	double                    lx, ly;
	char                     *markup;

	g_return_if_fail (GSD_IS_WACOM_OSD_BUTTON(osd_button));

	priv = osd_button->priv;

	cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);
	cairo_set_line_width (cr, LINE_WIDTH);

	actual_position = get_actual_position (priv->position, rotation);

	if (osd_button->priv->type == WACOM_TABLET_BUTTON_TYPE_HARDCODED)
		draw_circle (cr,
		             priv->area.x + priv->area.width / 2,
		             priv->area.y + priv->area.height / 2,
		             MIN (priv->area.width, priv->area.height) / 2);
	else
		draw_rounded_rectangle(cr,
		                       priv->area.x, priv->area.y,
		                       priv->area.width, priv->area.height);

	if (osd_button->priv->active)
		cairo_fill_preserve(cr);
	cairo_stroke (cr);

	/* Write the label */
	layout = pango_layout_new (pango_context);

	switch (actual_position) {
	case WACOM_TABLET_BUTTON_POS_TOP:
	case WACOM_TABLET_BUTTON_POS_BOTTOM:
		pango_layout_set_alignment (layout, PANGO_ALIGN_CENTER);
		break;
	case WACOM_TABLET_BUTTON_POS_RIGHT:
		pango_layout_set_alignment (layout, PANGO_ALIGN_RIGHT);
		break;
	default:
		pango_layout_set_alignment (layout, PANGO_ALIGN_LEFT);
		break;
	}
	markup = g_strdup_printf ("<span foreground=\"white\">%s</span>", priv->label);
	pango_layout_set_markup (layout, markup, -1);
	g_free (markup);

	pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
	switch (actual_position) {
	case WACOM_TABLET_BUTTON_POS_TOP:
		lx = priv->area.x + (priv->area.width - logical_rect.width) / 2 + logical_rect.x;
		ly = priv->area.y + (priv->area.height + LABEL_SPACING + logical_rect.y);
		break;
	case WACOM_TABLET_BUTTON_POS_BOTTOM:
		lx = priv->area.x + (priv->area.width - logical_rect.width) / 2 + logical_rect.x;
		ly = priv->area.y - (LABEL_SPACING + logical_rect.y + logical_rect.height);
		break;
	case WACOM_TABLET_BUTTON_POS_RIGHT:
		lx = priv->area.x - (LABEL_SPACING + logical_rect.x + logical_rect.width);
		ly = priv->area.y + (priv->area.height - logical_rect.height) / 2 + logical_rect.y;
		break;
	default:
		lx = priv->area.x + (priv->area.width + LABEL_SPACING + logical_rect.x);
		ly = priv->area.y + (priv->area.height - logical_rect.height) / 2 + logical_rect.y;
		break;
	}
	gtk_render_layout (style_context, cr, lx, ly, layout);

	g_object_unref (layout);
}

enum {
  PROP_OSD_WINDOW_0,
  PROP_OSD_WINDOW_MESSAGE,
  PROP_OSD_WINDOW_GSD_WACOM_DEVICE
};

#define GSD_WACOM_OSD_WINDOW_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), \
					     GSD_TYPE_WACOM_OSD_WINDOW, \
					     GsdWacomOSDWindowPrivate))

struct GsdWacomOSDWindowPrivate
{
	char                     *message;
	GsdWacomDevice           *pad;
	GsdWacomRotation          rotation;
	gint                      num_buttons[4]; /* How many button per side */
	GList                    *buttons;
};

static void     gsd_wacom_osd_window_class_init  (GsdWacomOSDWindowClass *klass);
static void     gsd_wacom_osd_window_init        (GsdWacomOSDWindow      *osd_window);
static void     gsd_wacom_osd_window_finalize    (GObject                *object);

G_DEFINE_TYPE (GsdWacomOSDWindow, gsd_wacom_osd_window, GTK_TYPE_WINDOW)

static void
gsd_wacom_osd_window_draw_message (GsdWacomOSDWindow   *osd_window,
				   GtkStyleContext     *style_context,
				   PangoContext        *pango_context,
				   cairo_t             *cr,
				   GtkAllocation       *allocation)
{
	PangoRectangle logical_rect;
	PangoLayout *layout;
	char *markup;
	double x;
	double y;

	layout = pango_layout_new (pango_context);
	pango_layout_set_alignment (layout, PANGO_ALIGN_CENTER);

	markup = g_strdup_printf ("<span foreground=\"white\">%s</span>", osd_window->priv->message);
	pango_layout_set_markup (layout, markup, -1);
	g_free (markup);

	pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
	x = (allocation->width - logical_rect.width) / 2 + logical_rect.x;
	y = (allocation->height - logical_rect.height) / 2 + logical_rect.y;

	gtk_render_layout (style_context, cr, x, y, layout);
	g_object_unref (layout);
}


static gboolean
gsd_wacom_osd_window_draw (GtkWidget *widget,
			   cairo_t   *cr)
{
	GsdWacomOSDWindow *osd_window = GSD_WACOM_OSD_WINDOW (widget);

	if (gtk_cairo_should_draw_window (cr, gtk_widget_get_window (widget))) {
		GtkAllocation        allocation;
		GtkStyleContext     *style_context;
		PangoContext        *pango_context;
		GList               *l;

		style_context = gtk_widget_get_style_context (widget);
		pango_context = gtk_widget_get_pango_context (widget);
		gtk_widget_get_allocation(widget, &allocation);

		cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
		cairo_paint (cr);
		cairo_set_operator (cr, CAIRO_OPERATOR_OVER);

		/* Draw all buttons */
		for (l = osd_window->priv->buttons; l != NULL; l = l->next) {
			GsdWacomOSDButton *osd_button = l->data;
			gsd_wacom_osd_button_draw (osd_button,
			                           style_context,
			                           pango_context,
			                           cr,
			                           &allocation,
			                           osd_window->priv->rotation);
		}

		/* Draw message */
		gsd_wacom_osd_window_draw_message (osd_window,
		                                   style_context,
		                                   pango_context,
		                                   cr,
		                                   &allocation);
	}

	return FALSE;
}

static void
gsd_wacom_osd_window_place_buttons (GsdWacomOSDWindow *osd_window,
				    GtkAllocation *allocation)
{
	gint              buttons[4] = { 0, 0, 0, 0 };
	GsdWacomRotation  rotation;
	gint              nw, nh, pw, ph;
	GList            *l;

	g_return_if_fail (GSD_IS_WACOM_OSD_WINDOW (osd_window));

	rotation = osd_window->priv->rotation;
	nw = MAX (osd_window->priv->num_buttons[get_actual_position (WACOM_TABLET_BUTTON_POS_TOP,    rotation)],
	          osd_window->priv->num_buttons[get_actual_position (WACOM_TABLET_BUTTON_POS_BOTTOM, rotation)]) + 3;
	nh = MAX (osd_window->priv->num_buttons[get_actual_position (WACOM_TABLET_BUTTON_POS_LEFT,   rotation)],
	          osd_window->priv->num_buttons[get_actual_position (WACOM_TABLET_BUTTON_POS_RIGHT,  rotation)]) + 3;

	pw = MIN (BUTTON_MAX_WIDTH, (allocation->width - (nw + 1) * BUTTON_SPACING) / nw);
	ph = pw * BUTTON_ASPECT;

	if ((ph + BUTTON_SPACING) * nh > allocation->height) {
		ph = (allocation->height - (nh + 1) * BUTTON_SPACING) / nh;
		pw = ph / BUTTON_ASPECT;
	}

	for (l = osd_window->priv->buttons; l != NULL; l = l->next) {
		GsdWacomTabletButtonPos  actual_position;
		GsdWacomOSDButton *osd_button = l->data;
		gint x, y, index, n_items, current;

		n_items = osd_window->priv->num_buttons[osd_button->priv->position];
		current = buttons[osd_button->priv->position];
		index = get_actual_index (osd_button->priv->position, rotation, n_items, current);
		actual_position = get_actual_position (osd_button->priv->position, rotation);

		gsd_wacom_osd_button_resize (osd_button, pw, ph);
		switch (actual_position) {
		case WACOM_TABLET_BUTTON_POS_LEFT:
			x = BUTTON_SPACING;
			y = (allocation->height - (n_items * (osd_button->priv->area.height + BUTTON_SPACING) - BUTTON_SPACING)) / 2
			    + index * (osd_button->priv->area.height + BUTTON_SPACING);
			break;
		case WACOM_TABLET_BUTTON_POS_RIGHT:
			x = allocation->width - (osd_button->priv->area.width + BUTTON_SPACING);
			y = (allocation->height - (n_items * (osd_button->priv->area.height + BUTTON_SPACING) - BUTTON_SPACING)) / 2
			    + index * (osd_button->priv->area.height + BUTTON_SPACING);
			break;
		case WACOM_TABLET_BUTTON_POS_TOP:
			x = (allocation->width - (n_items * (osd_button->priv->area.width + BUTTON_SPACING) - BUTTON_SPACING)) / 2
			    + index * (osd_button->priv->area.width + BUTTON_SPACING);
			y = BUTTON_SPACING;
			break;
		case WACOM_TABLET_BUTTON_POS_BOTTOM:
			x = (allocation->width - (n_items * (osd_button->priv->area.width + BUTTON_SPACING) - BUTTON_SPACING)) / 2
			    + index * (osd_button->priv->area.width + BUTTON_SPACING);
			y = allocation->height - (osd_button->priv->area.height + BUTTON_SPACING);
			break;
		case WACOM_TABLET_BUTTON_POS_UNDEF:
		default:
			g_assert_not_reached();
			break;
		}
		gsd_wacom_osd_button_move (osd_button, x, y);
		buttons[osd_button->priv->position]++;
	}
}

static gchar *
get_tablet_button_label (GsdWacomDevice       *device,
	                 GsdWacomTabletButton *button)
{
	gchar *str;
	guint keyval;
	GdkModifierType mask;

	g_return_val_if_fail (button, NULL);

	if (button->type == WACOM_TABLET_BUTTON_TYPE_HARDCODED) {
		gint mode;

		mode = gsd_wacom_device_get_current_mode (device, button->group_id);
		return g_strdup_printf (_("%s\nCurrent mode %d"), button->name, mode);
	}

	if (!button->settings)
		goto out;

	if (button->type == WACOM_TABLET_BUTTON_TYPE_NORMAL) {
		GsdWacomActionType type;
		gchar *name;

		type = g_settings_get_enum (button->settings, ACTION_TYPE_KEY);
		if (type == GSD_WACOM_ACTION_TYPE_NONE)
			return g_strdup (_("None"));

		if (type == GSD_WACOM_ACTION_TYPE_HELP)
			return g_strdup (_("Show On-Screen Help"));

		str = g_settings_get_string (button->settings, CUSTOM_ACTION_KEY);
		if (str == NULL || *str == '\0') {
			g_free (str);
			return g_strdup (_("None"));
		}

		gtk_accelerator_parse (str, &keyval, &mask);
		g_free (str);

		str = gtk_accelerator_get_label (keyval, mask);
		name = g_markup_printf_escaped ("Send Keystroke %s", str);
		g_free (str);

		return name;
	}

out:
	return g_strdup (button->name);
}

/*
 * Returns the rotation to apply a device to get a represntation relative to
 * the current rotation of the output.
 * (This function is _not_ the same as in gsd-wacom-manager.c)
 */
static GsdWacomRotation
display_relative_rotation (GsdWacomRotation device_rotation,
			   GsdWacomRotation output_rotation)
{
	GsdWacomRotation rotations[] = { GSD_WACOM_ROTATION_HALF,
	                                 GSD_WACOM_ROTATION_CW,
	                                 GSD_WACOM_ROTATION_NONE,
	                                 GSD_WACOM_ROTATION_CCW };
	guint i;

	if (device_rotation == output_rotation)
		return GSD_WACOM_ROTATION_NONE;

	if (output_rotation == GSD_WACOM_ROTATION_NONE)
		return device_rotation;

	for (i = 0; i < G_N_ELEMENTS (rotations); i++) {
		if (device_rotation == rotations[i])
			break;
	}

	if (output_rotation == GSD_WACOM_ROTATION_HALF)
		return rotations[(i + G_N_ELEMENTS (rotations) - 2) % G_N_ELEMENTS (rotations)];

	if (output_rotation == GSD_WACOM_ROTATION_CW)
		return rotations[(i + 1) % G_N_ELEMENTS (rotations)];

	if (output_rotation == GSD_WACOM_ROTATION_CCW)
		return rotations[(i + G_N_ELEMENTS (rotations) - 1) % G_N_ELEMENTS (rotations)];

	/* fallback */
	return GSD_WACOM_ROTATION_NONE;
}

void
gsd_wacom_osd_window_set_device (GsdWacomOSDWindow *osd_window,
				 GsdWacomDevice    *device)
{
	GtkAllocation     allocation;
	GsdWacomRotation  device_rotation;
	GsdWacomRotation  output_rotation;
	GSettings        *settings;
	gint              monitor;
	GdkRectangle      desktop;
	GdkScreen        *screen;
	GList            *list, *l;

	g_return_if_fail (GSD_IS_WACOM_OSD_WINDOW (osd_window));
	g_return_if_fail (GSD_IS_WACOM_DEVICE (device));

	/* Bind the device with the OSD window */
	if (osd_window->priv->pad)
		g_object_weak_unref (G_OBJECT(osd_window->priv->pad),
		                     (GWeakNotify) gtk_widget_destroy,
		                     osd_window);
	osd_window->priv->pad = device;
	g_object_weak_ref (G_OBJECT(osd_window->priv->pad),
	                   (GWeakNotify) gtk_widget_destroy,
	                   osd_window);

	/* Determine the monitor for that device */
	monitor = gsd_wacom_device_get_display_monitor (device);
	if (monitor < 0)
		monitor = 0;
	screen = gtk_window_get_screen (GTK_WINDOW (osd_window));
	if (screen == NULL)
		screen = gdk_screen_get_default ();
	gdk_screen_get_monitor_geometry (screen, monitor, &desktop);

	/* and place the OSD window accordingly */
	gtk_window_move (GTK_WINDOW (osd_window), desktop.x, desktop.y);
	gtk_window_set_default_size (GTK_WINDOW (osd_window), desktop.width, desktop.height);

	/* Capture current rotation, we do not update that later, OSD window is meant to be short lived */
	settings = gsd_wacom_device_get_settings (osd_window->priv->pad);
	device_rotation = g_settings_get_enum (settings, ROTATION_KEY);
	output_rotation = gsd_wacom_device_get_display_rotation (osd_window->priv->pad);
	osd_window->priv->rotation = display_relative_rotation (device_rotation, output_rotation);

	/* Create the new button hash table */
	if (osd_window->priv->buttons)
		g_list_free_full (osd_window->priv->buttons, g_object_unref);
	osd_window->priv->buttons = NULL;

	/* Create the buttons and compute the number of buttons on each side */
	list = gsd_wacom_device_get_buttons (device);
	for (l = list; l != NULL; l = l->next) {
		GsdWacomTabletButton *tablet_button = l->data;

		if (tablet_button->type == WACOM_TABLET_BUTTON_TYPE_NORMAL ||
		    tablet_button->type == WACOM_TABLET_BUTTON_TYPE_HARDCODED) {
			GsdWacomOSDButton *osd_button;
			gchar *str;

			/* For now, place all buttons at (0,0) we'll move them later */
			osd_button = gsd_wacom_osd_button_new (0, 0, 0, 0);
			str = get_tablet_button_label (device, tablet_button);
			gsd_wacom_osd_button_set_label (osd_button, str);
			g_free (str);
			gsd_wacom_osd_button_set_id (osd_button, tablet_button->id);
			gsd_wacom_osd_button_set_button_type (osd_button, tablet_button->type);
			gsd_wacom_osd_button_set_position (osd_button, tablet_button->pos);
			osd_window->priv->buttons = g_list_append (osd_window->priv->buttons, osd_button);

			osd_window->priv->num_buttons[tablet_button->pos]++;
		}
	}
	g_list_free (list);

	gtk_widget_get_allocation(GTK_WIDGET(osd_window), &allocation);
	gsd_wacom_osd_window_place_buttons (osd_window, &allocation);
}

GsdWacomDevice *
gsd_wacom_osd_window_get_device (GsdWacomOSDWindow *osd_window)
{
	g_return_val_if_fail (GSD_IS_WACOM_OSD_WINDOW (osd_window), NULL);

	return osd_window->priv->pad;
}

void
gsd_wacom_osd_window_set_message (GsdWacomOSDWindow *osd_window,
				  const gchar       *str)
{
	g_return_if_fail (GSD_IS_WACOM_OSD_WINDOW (osd_window));
	g_return_if_fail (str != NULL);

	g_free (osd_window->priv->message);
	osd_window->priv->message = g_strdup (str);
}

const char *
gsd_wacom_osd_window_get_message (GsdWacomOSDWindow *osd_window)
{
	g_return_val_if_fail (GSD_IS_WACOM_OSD_WINDOW (osd_window), NULL);

	return osd_window->priv->message;
}

static void
gsd_wacom_osd_window_set_property (GObject        *object,
				   guint           prop_id,
				   const GValue   *value,
				   GParamSpec     *pspec)
{
	GsdWacomOSDWindow *osd_window;

	osd_window = GSD_WACOM_OSD_WINDOW (object);

	switch (prop_id) {
	case PROP_OSD_WINDOW_MESSAGE:
		gsd_wacom_osd_window_set_message (osd_window, g_value_get_string (value));
		break;
	case PROP_OSD_WINDOW_GSD_WACOM_DEVICE:
		gsd_wacom_osd_window_set_device (osd_window, g_value_get_object (value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
gsd_wacom_osd_window_get_property (GObject        *object,
				   guint           prop_id,
				   GValue         *value,
				   GParamSpec     *pspec)
{
	GsdWacomOSDWindow *osd_window;

	osd_window = GSD_WACOM_OSD_WINDOW (object);

	switch (prop_id) {
	case PROP_OSD_WINDOW_MESSAGE:
		g_value_set_string (value, osd_window->priv->message);
		break;
	case PROP_OSD_WINDOW_GSD_WACOM_DEVICE:
		g_value_set_object (value, (GObject*) osd_window->priv->pad);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

void
gsd_wacom_osd_window_set_active (GsdWacomOSDWindow *osd_window,
				 gchar             *button_id,
				 gboolean           active)
{
	GdkWindow *win;
	GList *l;

	g_return_if_fail (GSD_IS_WACOM_OSD_WINDOW (osd_window));
	g_return_if_fail (button_id != NULL);

	for (l = osd_window->priv->buttons; l != NULL; l = l->next) {
		GsdWacomOSDButton *osd_button = l->data;
		if (g_strcmp0 (osd_button->priv->id, button_id) == 0) {
			gsd_wacom_osd_button_set_active (osd_button, active);
			win = gtk_widget_get_window(GTK_WIDGET(osd_window));
			gdk_window_invalidate_rect(win, &osd_button->priv->area, FALSE);
		}
	}
}

GtkWidget *
gsd_wacom_osd_window_new (GsdWacomDevice       *pad,
			  const gchar          *message)
{
	GsdWacomOSDWindow *osd_window;
	GdkWindow         *window;
	GdkRGBA            black;
	GdkCursor         *cursor;

	osd_window = GSD_WACOM_OSD_WINDOW (g_object_new (GSD_TYPE_WACOM_OSD_WINDOW,
	                                                 "wacom-device", pad,
	                                                 "message", message,
	                                                 NULL));

	gtk_widget_set_app_paintable (GTK_WIDGET (osd_window), TRUE);

	gdk_rgba_parse (&black, "rgb(0,0,0)");
	gtk_window_set_opacity (GTK_WINDOW (osd_window), WINDOW_OPACITY);

	gtk_widget_realize (GTK_WIDGET (osd_window));
	window = gtk_widget_get_window (GTK_WIDGET (osd_window));
	gdk_window_set_background_rgba (window, &black);

	cursor = gdk_cursor_new (GDK_BLANK_CURSOR);
	gdk_window_set_cursor (window, cursor);
	g_object_unref (cursor);

	gtk_widget_set_can_focus (GTK_WIDGET (osd_window), TRUE);
	gtk_window_set_focus_on_map (GTK_WINDOW (osd_window), TRUE);
	gtk_window_set_decorated (GTK_WINDOW (osd_window), FALSE);
	gtk_window_set_deletable (GTK_WINDOW (osd_window), FALSE);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW (osd_window), TRUE);
	gtk_window_set_skip_pager_hint (GTK_WINDOW (osd_window), TRUE);
	gtk_window_set_keep_above (GTK_WINDOW (osd_window), TRUE);
	gtk_window_fullscreen (GTK_WINDOW (osd_window));

	return GTK_WIDGET (osd_window);
}

static void
gsd_wacom_osd_window_size_allocate (GtkWidget     *widget,
				    GtkAllocation *allocation)
{
	/* Recompute buttons position */
	gsd_wacom_osd_window_place_buttons (GSD_WACOM_OSD_WINDOW (widget), allocation);
}

static void
gsd_wacom_osd_window_class_init (GsdWacomOSDWindowClass *klass)
{
	GObjectClass *gobject_class;
	GtkWidgetClass *widget_class;

	gobject_class = G_OBJECT_CLASS (klass);
	widget_class  = GTK_WIDGET_CLASS (klass);

	gobject_class->set_property = gsd_wacom_osd_window_set_property;
	gobject_class->get_property = gsd_wacom_osd_window_get_property;
	gobject_class->finalize     = gsd_wacom_osd_window_finalize;
	widget_class->draw          = gsd_wacom_osd_window_draw;
	widget_class->size_allocate = gsd_wacom_osd_window_size_allocate;

	g_object_class_install_property (gobject_class,
	                                 PROP_OSD_WINDOW_MESSAGE,
	                                 g_param_spec_string ("message",
	                                                      "Window message",
	                                                      "The message shown in the OSD window",
	                                                      "",
	                                                      G_PARAM_READWRITE));
	g_object_class_install_property (gobject_class,
	                                 PROP_OSD_WINDOW_GSD_WACOM_DEVICE,
	                                 g_param_spec_object ("wacom-device",
	                                                      "Wacom device",
	                                                      "The Wacom device represented by the OSD window",
	                                                      GSD_TYPE_WACOM_DEVICE,
	                                                      G_PARAM_READWRITE));

	g_type_class_add_private (klass, sizeof (GsdWacomOSDWindowPrivate));
}

static void
gsd_wacom_osd_window_init (GsdWacomOSDWindow *osd_window)
{
	osd_window->priv = GSD_WACOM_OSD_WINDOW_GET_PRIVATE (osd_window);
}

static void
gsd_wacom_osd_window_finalize (GObject *object)
{
	GsdWacomOSDWindow *osd_window;
	GsdWacomOSDWindowPrivate *priv;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GSD_IS_WACOM_OSD_WINDOW (object));

	osd_window = GSD_WACOM_OSD_WINDOW (object);

	g_return_if_fail (osd_window->priv != NULL);

	priv = osd_window->priv;

	g_free (priv->message);
	priv->message = NULL;

	if (priv->buttons) {
		g_list_free_full (priv->buttons, g_object_unref);
		priv->buttons = NULL;
	}

	G_OBJECT_CLASS (gsd_wacom_osd_window_parent_class)->finalize (object);
}
