/*
 * Copyright (C) 2012 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Author: Olivier Fourdan <ofourdan@redhat.com>
 *
 */

#include "config.h"

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <dirent.h>
#include <glib/gi18n.h>
#include "gsd-wacom-osd-window.h"

static gboolean fake_device = FALSE;
static gboolean option_debug = FALSE;

static GsdWacomDevice *
search_pad_device (void)
{
	GdkDeviceManager *mgr;
	GList *list, *l;

	mgr = gdk_display_get_device_manager (gdk_display_get_default ());
	list = gdk_device_manager_list_devices (mgr, GDK_DEVICE_TYPE_SLAVE);
	for (l = list; l ; l = l->next) {
		GsdWacomDevice *device;

		device = gsd_wacom_device_new (l->data);
		if (gsd_wacom_device_get_device_type (device) == WACOM_TYPE_PAD)
			return (device);
		g_object_unref (device);
	}
	g_list_free (list);

	return NULL;
}

static GsdWacomDevice *
create_fake_device (void)
{
	return gsd_wacom_device_create_fake (WACOM_TYPE_PAD,
					     "Wacom Cintiq 21UX2",
					     "Wacom Cintiq 21UX2 pad");
}

static gboolean
on_key_release_event(GtkWidget   *widget,
                     GdkEventKey *event,
                     gpointer     data)
{
	if (event->type != GDK_KEY_RELEASE)
		return FALSE;
	if (event->keyval != GDK_KEY_Escape)
		return FALSE;

	gtk_main_quit();

	return FALSE;
}

int main(int argc, char** argv)
{
	GtkWidget *widget;
	GError *error = NULL;
	GOptionContext *context;
	GsdWacomDevice *device;
	gchar *message;
	const GOptionEntry entries[] = {
		{ "fake", 'f', 0, G_OPTION_ARG_NONE, &fake_device, "Fake Wacom Cintiq 21UX2", NULL },
		{ "debug", 'd', 0, G_OPTION_ARG_NONE, &option_debug, "Debug output", NULL },
		{ NULL }
	};

	gtk_init (&argc, &argv);

	context = g_option_context_new ("- test functions");
	g_option_context_add_main_entries (context, entries, GETTEXT_PACKAGE);

	if (g_option_context_parse (context, &argc, &argv, &error) == FALSE) {
		g_print ("Option parsing failed: %s\n", error->message);
		return 1;
	}

	if (option_debug)
		g_setenv ("G_MESSAGES_DEBUG", "all", TRUE);

	if (fake_device == FALSE)
		device = search_pad_device ();
	else
		device = create_fake_device ();

	if (device == NULL) {
		g_print ("No pad device found, consider using --fake\n");
		return 1;
	}

	gtk_init(&argc, &argv);

	message = g_strdup_printf ("<big><b>%s</b></big>\n<i>Test OSD application</i>", gsd_wacom_device_get_name (device));
	widget = gsd_wacom_osd_window_new (device, message);
	g_free (message);

	g_signal_connect (widget, "key-release-event",
			  G_CALLBACK(on_key_release_event), NULL);
	g_signal_connect (widget, "delete-event",
			  G_CALLBACK (gtk_main_quit), NULL);
	g_signal_connect (widget, "unmap",
			  G_CALLBACK (gtk_main_quit), NULL);

	gtk_widget_show (widget);
	gtk_main ();

	return 0;
}
